"""Factories for the ``traces`` app."""
from factory import DjangoModelFactory, Sequence

from .. import models


class TraceFactory(DjangoModelFactory):
    FACTORY_FOR = models.Trace

    ip = Sequence(lambda x: '66.66.66.{}'.format(x))


class BlacklistIPFactory(DjangoModelFactory):
    FACTORY_FOR = models.BlacklistIP

    ip = Sequence(lambda x: '77.77.77.{}'.format(x))


class BlacklistUserAgentFactory(DjangoModelFactory):
    FACTORY_FOR = models.BlacklistUserAgent

    user_agent = Sequence(lambda x: 'user_agent_{}'.format(x))



"""Models for the ``traces`` app."""
from django.db import models
from django.conf import settings
from django.contrib.contenttypes import generic
from django.contrib.contenttypes.models import ContentType


class Trace(models.Model):
    """
    Model to track a user's view hits.

    :creation_date: Auto-added datetime of the model creation.
    :view_name: Name of the called view.
    :ip: IP of the visitor.
    :user_agent: User agent of the request.
    :session_key: Session key in the request.
    :user: User who called the website, if not anonymous.
    :view_object: Object of the called view.
    :hits: Hit count of the user/view combination.

    """
    creation_date = models.DateTimeField(
        auto_now_add=True,
        editable=False,
        verbose_name=('Creation date'),
    )

    view_name = models.CharField(
        max_length=50,
        verbose_name=('View name'),
    )

    ip = models.IPAddressField(
        verbose_name=('IP'),
    )

    session_key = models.CharField(
        max_length=40,
        verbose_name=('Session key'),
    )

    user_agent = models.CharField(
        max_length=255,
        verbose_name=('User agent'),
    )

    user = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        null=True, blank=True,
        verbose_name=('User'),
    )

    # GFK 'view_object'
    content_type = models.ForeignKey(ContentType, blank=True, null=True)
    object_id = models.PositiveIntegerField(blank=True, null=True)
    view_object = generic.GenericForeignKey('content_type', 'object_id')

    hits = models.PositiveIntegerField(
        default=1,
        verbose_name=('Hits'),
    )

    class Meta:
        ordering = ('-creation_date', )

    def __unicode__(self):
        return u'{} ({})'.format(self.view_name, self.ip)


class BlacklistIP(models.Model):
    """
    Model to contain information about unwanted IPs.

    :ip: IP to be blacklisted.
    :description: Optional description of this entry.

    """
    ip = models.IPAddressField(
        verbose_name=('IP'),
        unique=True,
    )

    description = models.TextField(
        max_length=1024,
        verbose_name=('Description'),
        blank=True,
    )

    class Meta:
        ordering = ('ip', )

    def __unicode__(self):
        return u'{}'.format(self.ip)


class BlacklistUserAgent(models.Model):
    """
    Model to contain information about unwanted user agents.

    :user_agent: User agent to be blacklisted.
    :description: Optional description of this entry.

    """
    user_agent = models.CharField(
        max_length=255,
        verbose_name=('User agent'),
        unique=True,
    )

    description = models.TextField(
        max_length=1024,
        verbose_name=('Description'),
        blank=True,
    )

    class Meta:
        ordering = ('user_agent', )

    def __unicode__(self):
        return u'{}'.format(self.user_agent)



"""Tests for the models of the ``traces`` app."""
from django.test import TestCase

from . import factories


class TraceTestCase(TestCase):
    def setUp(self):
        self.instance = factories.TraceFactory()

    def test_model(self):
        self.assertTrue(self.instance.pk)


class BlacklistIPTestCase(TestCase):
    def setUp(self):
        self.instance = factories.BlacklistIPFactory()

    def test_model(self):
        self.assertTrue(self.instance.pk)


class BlacklistUserAgentTestCase(TestCase):
    def setUp(self):
        self.instance = factories.BlacklistUserAgentFactory()

    def test_model(self):
        self.assertTrue(self.instance.pk)



"""Custom middlewares for the ``traces`` app."""
from django.conf import settings
from django.contrib.contenttypes.models import ContentType
from django.core.urlresolvers import resolve, Resolver404
from django.core.validators import ipv4_re

from .models import BlacklistIP, BlacklistUserAgent, Trace


def get_ip(request):
    """
    Retrieves the remote IP address from the request data.  If the user is
    behind a proxy, they may have a comma-separated list of IP addresses, so
    we need to account for that.  In such a case, only the first IP in the
    list will be retrieved.  Also, some hosts that use a proxy will put the
    REMOTE_ADDR into HTTP_X_FORWARDED_FOR.  This will handle pulling back the
    IP from the proper place.

    """
    ip = request.META.get('HTTP_X_FORWARDED_FOR',
                          request.META.get('REMOTE_ADDR', '127.0.0.1'))
    ip = ipv4_re.match(ip)
    if not ip:
        return None
    return ip.group(0)


class TracesMiddleware(object):
    """
    Middleware that checks if the url name matches a traced view.

    If the trace should be tracked, a new hit is counted.

    """
    def process_response(self, request, response):
        try:
            view_name = resolve(request.path_info).url_name
        except Resolver404:
            return response
        if view_name in getattr(settings, 'TRACED_VIEWS', []):
            ip = get_ip(request)
            if not ip:
                return response
            user_agent = request.META.get('HTTP_USER_AGENT', '')[:255]
            user = request.user
            if hasattr(response, 'context_data') and response.context_data:
                view_object = response.context_data.get('object')
            else:
                view_object = None

            # Check IP and user agent against blacklist
            if (BlacklistIP.objects.filter(ip__exact=ip)
                    or BlacklistUserAgent.objects.filter(
                        user_agent__exact=user_agent)):
                return response

            # Create session key, if unexistant
            if not request.session.session_key:
                request.session.modified = True
                request.session.save()

            trace_kwargs = {
                'view_name': view_name,
                'ip': ip,
                'user_agent': user_agent,
                'session_key': request.session.session_key,
            }
            filter_kwargs = {'view_name': view_name}
            if view_object:
                trace_kwargs['view_object'] = view_object
                filter_kwargs['content_type'] = \
                    ContentType.objects.get_for_model(view_object)
                filter_kwargs['object_id'] = view_object.pk

            # Check if trace exists
            if user.is_authenticated():
                filter_kwargs['user'] = user
                try:
                    trace = Trace.objects.get(**filter_kwargs)
                except Trace.DoesNotExist:
                    trace_kwargs['user'] = user
                    Trace.objects.create(**trace_kwargs)
                    return response
                trace.hits += 1
                trace.save()
                return response
            filter_kwargs['session_key'] = request.session.session_key
            try:
                trace = Trace.objects.get(**filter_kwargs)
            except Trace.DoesNotExist:
                Trace.objects.create(**trace_kwargs)
                return response
            trace.hits += 1
            trace.save()
            return response
        return response



"""Test for the template tags of the ``traces`` app."""
from django.test import TestCase

from django_libs.tests.factories import UserFactory
from mock import Mock

from ..templatetags import traces_tags
from . import factories


class GetViewHitsTestCase(TestCase):
    """Tests for the ``get_view_hits`` template tag."""
    longMessage = True

    def test_tag(self):
        self.assertEqual(traces_tags.get_view_hits({}, 'test_view'), 0, msg=(
            'Should return one hit.'))

        with self.settings(TRACED_VIEWS=['view_with_hits']):
            factories.TraceFactory(view_name='view_with_hits', hits='4')
            self.assertEqual(
                traces_tags.get_view_hits({}, 'view_with_hits'), 4,
                msg=('Should return five hits.'))

        with self.settings(TRACED_VIEWS=[]):
            self.assertIsNone(traces_tags.get_view_hits({}, 'test_view'))

            request = Mock()
            request.path_info = '/inexistant-view/'
            context = {'request': request}
            self.assertIsNone(traces_tags.get_view_hits(context))

        with self.settings(TRACED_VIEWS=['view_with_an_object']):
            context = {'object': UserFactory()}
            self.assertEqual(
                traces_tags.get_view_hits(context, 'view_with_an_object'), 0,
                msg=('Should return five hits.'))



"""Admin registration for the ``traces`` app."""
from django.contrib import admin

from . import models


admin.site.register(models.BlacklistIP)
admin.site.register(models.BlacklistUserAgent)
admin.site.register(models.Trace)



"""URLs to run the tests."""
from django.conf.urls import patterns, include, url
from django.contrib import admin
from django.views.generic import DetailView, TemplateView

from test_app.models import DummyModel


admin.autodiscover()

urlpatterns = patterns(
    '',
    url(r'^admin/', include(admin.site.urls)),
    url(r'^(?P<pk>\d+)/$',
        DetailView.as_view(model=DummyModel, template_name='base.html'),
        name='test_dummy_view'),
    url(r'^$', TemplateView.as_view(template_name='base.html'),
        name='test_view'),
)



"""Tests for the middlewares of the ``traces`` app."""
from django.contrib.auth.models import AnonymousUser
from django.test import TestCase

from django_libs.tests.factories import UserFactory
from mock import Mock

from factories import BlacklistIPFactory
from ..middleware import TracesMiddleware
from ..models import Trace


class TraceMiddlewareTestCase(TestCase):
    longMessage = True

    def setUp(self):
        self.middleware = TracesMiddleware()
        self.request = Mock()
        self.request.user = AnonymousUser()
        self.request.path_info = '/'
        self.request.session.session_key = 'foobar'
        self.request.resolver_match.url_name = 'test_view'
        self.request.META = {'HTTP_USER_AGENT': ''}
        self.response = Mock()
        self.response.context_data = None

    def test_untraced_view(self):
        with self.settings(TRACED_VIEWS=[]):
            self.assertTrue(
                self.middleware.process_response(self.request, self.response))
            self.assertEqual(Trace.objects.count(), 0, msg=(
                'No trace should have been created.'))

    def test_traced_view(self):
        with self.settings(TRACED_VIEWS=['test_view']):
            # Anonymous user
            self.assertTrue(
                self.middleware.process_response(self.request, self.response))
            self.assertEqual(Trace.objects.count(), 1, msg=(
                'A new trace should have been created.'))

            self.assertTrue(
                self.middleware.process_response(self.request, self.response))
            self.assertEqual(Trace.objects.count(), 1, msg=(
                'No new trace should have been created.'))
            self.assertEqual(Trace.objects.all()[0].hits, 2, msg=(
                'Hits should have been increased.'))

            # Blacklisted
            BlacklistIPFactory(ip=Trace.objects.get().ip)
            self.assertTrue(
                self.middleware.process_response(self.request, self.response))
            self.assertEqual(Trace.objects.count(), 1, msg=(
                'No new trace should have been created.'))

            self.request.session.session_key = ''
            self.request.META['HTTP_X_FORWARDED_FOR'] = '1.1.1.1'
            self.assertTrue(
                self.middleware.process_response(self.request, self.response))

            # Logged in user
            self.request.user = UserFactory()
            self.assertTrue(
                self.middleware.process_response(self.request, self.response))
            self.assertEqual(Trace.objects.count(), 3, msg=(
                'A new trace should have been created.'))

            self.assertTrue(
                self.middleware.process_response(self.request, self.response))
            self.assertEqual(Trace.objects.count(), 3, msg=(
                'No new trace should have been created.'))
            self.assertEqual(Trace.objects.all()[0].hits, 2, msg=(
                'Hits should have been increased.'))

            # View object
            self.response.context_data = {'object': UserFactory()}
            self.request.resolver_match.url_name = 'test_model_view'
            self.assertTrue(
                self.middleware.process_response(self.request, self.response))
            self.assertEqual(Trace.objects.count(), 4, msg=(
                'A new trace should have been created.'))

            # Invalid IP
            self.request.META['HTTP_X_FORWARDED_FOR'] = '1.1.1.1.1.1.1'
            self.assertTrue(
                self.middleware.process_response(self.request, self.response))
            self.assertEqual(Trace.objects.count(), 4, msg=(
                'No new trace should have been created.'))

            # Invalid URL or missing view name
            self.request.path_info = '/inexistant-view/'
            self.assertTrue(
                self.middleware.process_response(self.request, self.response))
            self.assertEqual(Trace.objects.count(), 4, msg=(
                'No new trace should have been created.'))



"""Template tags for the ``traces`` app."""
from django.conf import settings
from django.contrib.contenttypes.models import ContentType
from django.core.urlresolvers import resolve, Resolver404
from django.db.models import Sum
from django.template import Library

from ..models import Trace


register = Library()


@register.assignment_tag(takes_context=True)
def get_view_hits(context, view_name='', view_object=None):
    """Simply returns the amount of hits for this view."""
    try:
        view_name = view_name or resolve(context['request'].path_info).url_name
    except Resolver404:
        return None
    if view_name not in getattr(settings, 'TRACED_VIEWS', []):
        # If it's not a traced view return None
        return None
    filter_kwargs = {'view_name': view_name}
    view_object = view_object or context.get('object')
    if view_object:
        filter_kwargs['content_type'] = \
            ContentType.objects.get_for_model(view_object)
        filter_kwargs['object_id'] = view_object.pk
    traces = Trace.objects.filter(**filter_kwargs)
    if traces:
        # If there are saved traces aggregate their hits
        # Since the hit is updated with the view response (and not yet saved)
        # we can already add a 1
        return traces.aggregate(hits=Sum('hits'))['hits']
    # If it's a traced view we know that a new trace will be created with
    # the response, so we return one hit
    return 0
